"""Cosmological calculations and methods.

This file primarily provides the `Cosmology` class, which extends the functionality of astropy's
`FlatLambdaCDM` class to include additional utility functions.  The `Cosmology` class here is
initialized with WMAP9 parameters from [WMAP9]_.

References
----------
* [Hogg1999]_ Hogg 1999.
* [WMAP9]_ Hinshaw et al. 2013.

"""

from builtins import super

import astropy as ap
import astropy.cosmology  # noqa
import scipy as sp
import scipy.interpolate  # noqa
import numpy as np

from . import SPLC

# These are WMAP9 parameters, see [WMAP9], Table 3, WMAP+BAO+H0
Omega0 = 0.2880                #: Matter density parameter "Om0"
OmegaBaryon = 0.0472           #: Baryon density parameter "Ob0"
HubbleParam = 0.6933           #: Hubble Parameter as H0/[100 km/s/Mpc], i.e. 0.69 instead of 69
# Hubble0 = HubbleParam * 100.0  # NOTE: this *cannot* be named `H0` or `_H0` --- conflicts with astropy internals

# Define redshift grid: `_Z_GRID` defines the course grid points, between which `_GRID_SIZE` points
# are added inbetween.  i.e. if ``N = len(_Z_GRID)`` then there will be `N * _GRID_SIZE + 1` total
# grid points
# NOTE: z=0.0 is added automatically
_Z_GRID = [1e4, 100.0, 10.0, 4.0, 2.0, 1.0, 0.5, 0.1, 0.01]
_GRID_SIZE_DEF = 100


class Cosmology(ap.cosmology.FlatLambdaCDM):
    """Class for calculating (and inverting) cosmological distance measures.

    Based on the `astropy.cosmology.Cosmology` methods and classes.

    The base class provides methods to convert from redshift to desired cosmological parameters,
    e.g. `luminosity_distance(z)`.  These methods are also used to construct a numerical grid of
    values as a function of redshift.  This numerical grid is then used to interpolate backwards
    from cosmological parameters (e.g. distance measures) back to redshift.

    """

    def __init__(self, h=None, H0=None, Om0=None, Ob0=None, size=None, **kwargs):
        """
        """
        # ---- Set Defaults
        if (H0 is not None) and (h is not None):
            raise ValueError("Both `h` and `H0` cannot both be given!")
        if H0 is None:
            if h is None:
                h = HubbleParam
            H0 = h * 100.0  # Hubble Parameter h = H0/[100 km/s/Mpc], i.e. 0.69 instead of 69 km/s/Mpc
        if Om0 is None:
            Om0 = Omega0
        if Ob0 is None:
            Ob0 = OmegaBaryon
        kw = dict(H0=H0, Om0=Om0, Ob0=Ob0)
        kwargs.update(kw)
        if size is None:
            size = _GRID_SIZE_DEF
        self._size = size

        # Initialize parent class
        super().__init__(**kwargs)

        # Create grids for interpolations
        # -------------------------------
        #    Create a grid in redshift at which functions will be evaluated
        zgrid = self._init_interp_grid(_Z_GRID, self._size)
        self._grid_z = zgrid
        self._sort_z = np.argsort(zgrid)
        self._grid_a = self.z_to_a(zgrid)
        self._sort_a = np.argsort(self._grid_a)
        # Calculate corresponding values in desired parameters
        #    Ages in seconds
        self._grid_age = self.age(zgrid).cgs.value
        self._sort_age = np.argsort(self._grid_age)
        self._grid_lbk = self.lookback_time(zgrid).cgs.value
        self._sort_lbk = np.argsort(self._grid_lbk)
        #    Comoving distances in centimeters
        self._grid_dcom = self.comoving_distance(zgrid).cgs.value
        self._sort_dcom = np.argsort(self._grid_dcom)
        #    Luminosity distances in centimeters
        self._grid_dlum = self.luminosity_distance(zgrid).cgs.value
        self._sort_dlum = np.argsort(self._grid_dlum)
        return

    def __str__(self):
        """Return a string description of this instance.
        """
        rstr = "{} :: H0 = {:.8f}, Om0 = {:.8f}, Ob0 = {:.8f}".format(
            # __class__, self.Hubble0, self.Omega0, self.OmegaBaryon
            __class__, self.H0, self.Om0, self.Ob0
        )
        return rstr

    @staticmethod
    def _init_interp_grid(z_pnts, num_pnts):
        """Create a grid in redshift for interpolation.

        Arguments
        ---------
        z_pnts : (N,) array_like of scalar
            Monotonically decreasing redshift values greater than zero.  Inbetween each pair of
            values, `num_pnts` points are evenly added in log-space.  Also, `num_pnts` are added
            from the minimum given value down to redshift `0.0`.
        num_pnts : int
            Number of grid points to insert between each pair of values in `z_pnts`.

        Returns
        -------
        zgrid : (M,) array of scalar
            The total length of the final array will be ``N * num_pnts + 1``; that is, `num_pnts`
            for each value given in `z_pnts`, plus redshift `0.0` at the end.

        """
        # Make sure grid is monotonically decreasing
        if not np.all(np.diff(z_pnts) < 0.0):
            err_str = "Non-monotonic z_grid = {}".format(z_pnts)
            raise ValueError(err_str)

        z0 = z_pnts[0]
        zgrid = None
        # Create a log-spacing of points between each pair of values in the given points
        for z1 in z_pnts[1:]:
            temp = np.logspace(*np.log10([z0, z1]), num=num_pnts, endpoint=False)
            if zgrid is None:
                zgrid = temp
            else:
                zgrid = np.append(zgrid, temp)
            z0 = z1

        # Linearly space from the last point up to `0.0` (cant reach `0.0` with log-spacing)
        zgrid = np.append(zgrid, np.linspace(z0, 0.0, num=num_pnts))
        return zgrid

    @staticmethod
    def a_to_z(sf):
        """Convert from scale-factor to redshift.

        :math:`z = (1/a) - 1`

        Arguments
        ---------
        sf : array_like
            Scale-factor (:math:`a`) of the universe.

        Returns
        -------
        float or array
            Redshift at the input scale-factors.

        """
        sf = np.asarray(sf)
        # NOTE: this does not check for `nan`
        if not np.all((sf > 0.0) & (sf <= 1.0)):
            raise ValueError("Scale-factor must be [0.0, 1.0]")

        return (1.0/sf) - 1.0

    @staticmethod
    def z_to_a(redz):
        """Convert from redshift to scale-factor.

        :math:`a = 1.0 / (1 + z)`

        Arguments
        ---------
        redz : array_like
            Redshift(s) (:math:`z`) of the universe.

        Returns
        -------
        float or array
            Scale factor(s) at the inpur redshift(s).

        """
        redz = np.asarray(redz)
        # NOTE: this does not check for `nan`
        if not np.all(redz >= 0.0):
            raise ValueError("Redshift must be [0.0, +inf)")

        return 1.0/(redz + 1.0)

    @staticmethod
    def _interp(vals, xx, yy, inds=None):
        """Interpolate to the target locations.
        """
        if inds is None:
            inds = np.argsort(xx)
        # PchipInterpolate guarantees monotonicity with higher order
        arrs = sp.interpolate.PchipInterpolator(xx[inds], yy[inds], extrapolate=False)(vals)
        return arrs

    def _get_grid(self):
        """Return an array of the grid of interpolation points for each cosmological parameter.

        Returns
        -------
        grid : (N,M,) of scalar
            Grid of `N` values for each of `M` cosmological parameters.
            For example, the redshift values are `grid[:, 0]`.
        names : (M,) str
            The names of each cosmological parameter in the grid.
        units : (M,) str
            The units used for each cosmological parameter in the grid.

        """
        _var_keys = ["_grid_z", "_grid_a", "_grid_dcom", "_grid_dlum", "_grid_lbk", "_grid_age"]
        _var_names = ["z", "a", "dc", "dl", "tl", "ta"]
        _var_types = [None, None, ['cm', 'Mpc'], ['cm', 'Mpc'], ['s', 'Gyr'], ['s', 'Gyr']]

        ngrid = self._grid_z.size
        npars = len(_var_keys)
        grid = np.zeros((ngrid, npars))
        names = []
        units = []
        for ii, (vk, un) in enumerate(zip(_var_keys, _var_types)):
            vals = getattr(self, vk)
            nam = _var_names[ii]
            # Convert units if desired
            if un is not None:
                vals = ap.units.Quantity(vals, unit=un[0]).to(un[1]).value
                # nam += '[' + un[1] + ']'
                units.append(un[1])
            else:
                units.append('-')
            grid[:, ii] = vals
            names.append(nam)

        return grid, names, units

    # ==== a to distance measures ====

    def a_to_dcom(self, scafa):
        """Get the comoving distance [cm] at the given scale factor(s).

        Parameters
        ----------
        scafa : array_like,
            Scale-factor ($a$) describing the distance/time of the universe. Unitless in (0.0, 1.0].

        Returns
        -------
        dc : array_like
            Comoving distance to the given scale-factors.  Units of [cm].

        """
        dc = self._interp(scafa, self._grid_a, self._grid_dcom, self._sort_a)
        return dc

    def a_to_dlum(self, scafa):
        """Get the luminosity distance [cm] at the given scale factor(s).

        Parameters
        ----------
        scafa : array_like,
            Scale-factor ($a$) describing the distance/time of the universe. Unitless in (0.0, 1.0].

        Returns
        -------
        dl : array_like
            Luminosity distance to the given scale-factors.  Units of [cm].

        """
        dl = self._interp(scafa, self._grid_a, self._grid_dlum, self._sort_a)
        return dl

    def a_to_tage(self, sca):
        """Get the age of the universe [seconds] at the given scale factor(s).

        Parameters
        ----------
        scafa : array_like,
            Scale-factor ($a$) describing the distance/time of the universe. Unitless in (0.0, 1.0].

        Returns
        -------
        tage : array_like
            Age of the universe at the given scale-factors.  Units of [sec].

        """
        tage = self._interp(sca, self._grid_a, self._grid_age, self._sort_a)
        return tage

    def a_to_tlbk(self, scafa):
        """Get the lookback time [seconds] at the given scale factor(s).

        Parameters
        ----------
        scafa : array_like,
            Scale-factor ($a$) describing the distance/time of the universe.

        Returns
        -------
        tlook : array_like
            Lookback time to the universe at the given scale-factors.  Units of [sec].

        """
        tlook = self._interp(scafa, self._grid_a, self._grid_lbk, self._sort_a)
        return tlook

    # ==== z to distance measures ====

    def z_to_dcom(self, redz):
        """Get the comoving distance [cm] at the given redshift(s).

        Parameters
        ----------
        redz : array_like,
            Redshift ($z$) describing the distance/time of the universe. Unitless in [0.0, inf).

        Returns
        -------
        dc : array_like
            Comoving distance to the given redshift(s).  Units of [cm].

        """
        dc = self._interp(redz, self._grid_z, self._grid_dcom, self._sort_z)
        return dc

    def z_to_dlum(self, redz):
        """Get the luminosity distance [cm] at the given redshift(s).

        Parameters
        ----------
        redz : array_like,
            Redshift ($z$) describing the distance/time of the universe. Unitless in [0.0, inf).

        Returns
        -------
        dl : array_like
            Luminosity distance to the given redshift(s).  Units of [cm].

        """
        dl = self._interp(redz, self._grid_z, self._grid_dlum, self._sort_z)
        return dl

    def z_to_tage(self, redz):
        """Get the age of the universe [seconds] at the given redshift(s).

        Parameters
        ----------
        redz : array_like,
            Redshift ($z$) describing the distance/time of the universe. Unitless in [0.0, inf).

        Returns
        -------
        tage : array_like
            Age of the universe at the given redshift(s).  Units of [sec].

        """
        tage = self._interp(redz, self._grid_z, self._grid_age, self._sort_z)
        return tage

    def z_to_tlbk(self, redz):
        """Get the lookback time [seconds] at the given redshift(s).

        Parameters
        ----------
        redz : array_like,
            Redshift ($z$) describing the distance/time of the universe. Unitless in [0.0, inf).

        Returns
        -------
        tlook : array_like
            Lookback time to the universe at the given redshift(s).  Units of [sec].

        """
        zz = self._interp(redz, self._grid_z, self._grid_lbk, self._sort_z)
        return zz

    # ==== distance measures to a ====

    def dcom_to_a(self, dc):
        """Convert from comoving distance [cm] to scale-factor.

        Parameters
        ----------
        dc : array_like
            Comoving distance.  Units of [cm].

        Returns
        -------
        scafa : array_like,
            Scale-factor ($a$) describing the distance/time of the universe. Unitless in (0.0, 1.0].

        """
        scafa = self._interp(dc, self._grid_dcom, self._grid_a, self._sort_dcom)
        return scafa

    def dlum_to_a(self, dl):
        """Convert from luminosity distance [cm] to scale-factor.

        Parameters
        ----------
        dl : array_like
            Luminosity distance.  Units of [cm].

        Returns
        -------
        scafa : array_like,
            Scale-factor ($a$) describing the distance/time of the universe. Unitless in (0.0, 1.0].

        """
        scafa = self._interp(dl, self._grid_dlum, self._grid_a, self._sort_dlum)
        return scafa

    def tage_to_a(self, tage):
        """Convert from age of the universe [seconds] to scale-factor.

        Parameters
        ----------
        tage : array_like
            Age of the universe.  Units of [sec].

        Returns
        -------
        scafa : array_like,
            Scale-factor ($a$) describing the distance/time of the universe. Unitless in (0.0, 1.0].

        """
        scafa = self._interp(tage, self._grid_age, self._grid_a, self._sort_age)
        return scafa

    def tlbk_to_a(self, tlook):
        """Convert from lookback time [seconds] to scale-factor.

        Parameters
        ----------
        tlook : array_like
            Lookback time of the universe.  Units of [sec].

        Returns
        -------
        scafa : array_like,
            Scale-factor ($a$) describing the distance/time of the universe. Unitless in (0.0, 1.0].

        """
        scafa = self._interp(tlook, self._grid_lbk, self._grid_a, self._sort_lbk)
        return scafa

    # ==== distance measures to z ====

    def dcom_to_z(self, dc):
        """Convert from comoving distance [cm] to redshift.

        Parameters
        ----------
        dc : array_like
            Comoving distance.  Units of [cm].

        Returns
        -------
        redz : array_like,
            Redshift ($z$) describing the distance/time of the universe. Unitless in [0.0, inf).

        """
        zz = self._interp(dc, self._grid_dcom, self._grid_z, self._sort_dcom)
        return zz

    def dlum_to_z(self, dl):
        """Convert from luminosity distance [cm] to redshift.

        Parameters
        ----------
        dl : array_like
            Luminosity distance.  Units of [cm].

        Returns
        -------
        redz : array_like,
            Redshift ($z$) describing the distance/time of the universe. Unitless in [0.0, inf).

        """
        zz = self._interp(dl, self._grid_dlum, self._grid_z, self._sort_dlum)
        return zz

    def tage_to_z(self, age):
        """Convert from age of the universe [seconds] to redshift.

        Parameters
        ----------
        tage : array_like
            Age of the universe.  Units of [sec].

        Returns
        -------
        redz : array_like,
            Redshift ($z$) describing the distance/time of the universe. Unitless in [0.0, inf).

        """
        zz = self._interp(age, self._grid_age, self._grid_z, self._sort_age)
        return zz

    def tlbk_to_z(self, tlook):
        """Convert from lookback time [seconds] to redshift.

        Parameters
        ----------
        tlook : array_like
            Lookback time of the universe.  Units of [sec].

        Returns
        -------
        redz : array_like,
            Redshift ($z$) describing the distance/time of the universe. Unitless in [0.0, inf).

        """
        zz = self._interp(tlook, self._grid_lbk, self._grid_z, self._sort_lbk)
        return zz

    # ==== other cosmological measures ====

    def dVcdz(self, zz, cgs=True):
        """Differential comoving volume of the universe.

        From [Hogg1999]_ Eq. 28.

        Arguments
        ---------
        zz : array_like
            Target redshifts (:math:`z`).
        cgs : bool
            True: Return float(s) with the answer in CGS units [cm^3].
            False: Return them as astropy `Quantity` object, defaulting to [Mpc^3].

        Returns
        -------
        retval : float or array, [cm^3] or [Mpc^3]
            Differential volume of the universe at the given redshift(s).
            For the returned units, see `cgs` argument.

        """
        # This is 4*pi*D_h
        efac = self.efunc(zz)
        if cgs:
            retval = (4.0*np.pi*SPLC/self.H(0.0).cgs.value)
            com_dist = self.comoving_distance(zz).cgs.value
        else:
            retval = (4.0*np.pi*ap.constants.c/self.H(0.0)).decompose()
            com_dist = self.comoving_distance(zz)

        retval = retval * np.square(com_dist) / efac
        # Force `astropy` to simplify units, otherwise it gives `m * Mpc^2` units
        if not cgs:
            retval = retval.to('Mpc3')

        return retval

    def dtdz(self, zz):
        """Differential lookback time of the Universe.

        From [Hogg1999]_ Eq. 30

        Returns
        -------
        retval : float or array, units of [sec]
            dt/dz at the given redshift(s).

        """
        zz = np.asarray(zz)
        efac = self.efunc(zz)
        time_hub = self.hubble_time.to('s').value

        retval = time_hub / (1.0 + zz) / efac
        return retval



"""Interface to the cosmopy package via either command-line or python API.
"""
import argparse
import os
import sys

import click

import numpy as np
import astropy as ap

from . cosmology import Cosmology
from . import PC, ARCSEC, U_SEC, U_CM

_RESULTS_FUNCS = ['luminosity_distance', 'comoving_distance', 'lookback_time', 'age']

_RESULTS_FUNC_PARS = ['dl', 'dc', 'tl', 'ta']
_RESULTS_PARS = ['z', 'a'] + _RESULTS_FUNC_PARS

_COLOR_BASE = "magenta"  # "cyan"
_COLOR_CONV = "green"  # "yellow"
_COLOR_NAME = "red"

# Enable imperial units (e.g. 'miles')
ap.units.imperial.enable()


def calc_basic(cosmo, sets):
    """Given a `Cosmology` instance and input arguments, calculate basic cosmological values.

    Parameters
    ----------
    cosmo : `Cosmology`
        Class for calculating cosmological parameters, based on `astropy.cosmology` classes.
    sets : `argparse.Namespace`
        Namespace containing input parameters.
        One of its parameters, {`a`, `z`, `dc`, `dl`, `tl`, `ta`} must be non-None.

    Returns
    -------
    results : `Results` namedtuple
        Namedtuple of basic cosmological parameter results which can be processed into outputs.

    """
    scale = None

    results = {kk: None for kk in _RESULTS_PARS}

    # Redshift
    if sets.z is not None:
        redz = parse_input(sets.z)

    # Scalefactor
    elif sets.a is not None:
        scale = parse_input(sets.a)
        redz = cosmo.a_to_z(scale)

    # Age of the Universe
    elif sets.ta is not None:
        tage = parse_input(sets.ta, U_SEC)
        redz = cosmo.tage_to_z(tage.cgs.value)
        results['ta'] = tage

    # Loockback Time
    elif sets.tl is not None:
        tlbk = parse_input(sets.tl, U_SEC)
        redz = cosmo.tlbk_to_z(tlbk.cgs.value)
        results['tl'] = tlbk

    # Comoving Distance
    elif sets.dc is not None:
        dcom = parse_input(sets.dc, U_CM)
        redz = cosmo.dcom_to_z(dcom.cgs.value)
        # Calculate luminosity-distance manually
        dlum = dcom * (1.0 + redz)
        results['dc'] = dcom
        results['dl'] = dlum

    # Luminosity Distance
    elif sets.dl is not None:
        dlum = parse_input(sets.dl, U_CM)
        redz = cosmo.dlum_to_z(dlum.cgs.value)
        # Calculate comoving-distance manually
        dcom = dlum / (1.0 + redz)
        results['dc'] = dcom
        results['dl'] = dlum

    # No arguments set, raise error
    else:
        print("__main__.calc()")
        print("\t`sets` = '{}'".format(sets))
        raise ValueError("No input given!")

    # Calculate scale-factor if needed
    scale = cosmo.z_to_a(redz) if scale is None else scale
    results['a'] = scale
    results['z'] = redz

    # Calculate the values not already set
    for pp, ff in zip(_RESULTS_FUNC_PARS, _RESULTS_FUNCS):
        if results[pp] is None:
            func = getattr(cosmo, ff)
            results[pp] = func(redz)

    return results


def calc_derived(results):
    """Calculate derived cosmological parameters based on the standard ones.

    Adds:
        - 'da': the Angular diameter distance
        - 'arc': the physical size corresponding to one arcsecond
        - 'dm': the distance modulus

    Arguments
    ---------
    results : dict
        Dictionary of standard cosmological parameters (e.g. dl, z, tl, etc)

    Returns
    -------
    results : dict
        Dictionary of parameters with derived quantities added.

    """
    dang = results['dc']/(1.0 + results['z'])
    darc = dang*ARCSEC
    dist_mod = 5.0*np.log10(results['dl'].cgs.value/(10.0*PC))

    results['da'] = dang
    results['arc'] = darc
    results['dm'] = dist_mod

    return results


def output_print(results, print_output=True):
    """Format and print the given cosmological parameters to stdout.

    Arguments
    ---------
    results : `Results` namedtuple
        Cosmological parameters to output.
    print_output : bool
        Print output to stdout in addition to returning dictionary of values

    Returns
    -------
    retvals : dict
        Dictionary of cosmological parameter outputs in formatted str form.

    """

    keys = ['z', 'a', 'dc', 'dl',
            'da', 'arc', 'tl',
            'ta', 'dm']
    symbs = ['z', 'a', 'D_c', 'D_L',
             'D_A', 'Arcsec', 'T_lb',
             'T_a', 'DM']
    types = [None, None, 'Mpc', 'Mpc',
             'Mpc', 'pc', 'Gyr',
             'Gyr', None]
    names = ['Redshift', 'Scale-factor', 'Comoving Distance', 'Luminosity Distance',
             'Angular Diameter Distance', 'Arcsecond Scale', 'Lookback Time',
             'Age of the Universe', 'Distance Modulus']

    retvals = {}
    printvals = []
    outs = []
    for kk, ss, tt, nn in zip(keys, symbs, types, names):
        vv = results[kk]
        try:
            vv = vv.item()
        except AttributeError:
            pass
        v_std = vv if tt is None else vv.to(tt)
        base = "{:>10s} = {:.4f}".format(ss, v_std)
        # try:
        #     base = "{:>10s} = {:.4f}".format(ss, v_std)
        # except Exception:
        #     print("Failed constructing `base` string on '{}' ({})".format(nn, ss))
        #     print("  Value = '{}', type = '{}'".format(v_std, type(v_std)))
        #     raise

        try:
            v_cgs = vv.cgs
            conv = " ~ {:.4e}".format(v_cgs)
        except Exception:
            conv = ""

        rstr = "{base:30s}{conv:20s} : {name}".format(base=base, conv=conv, name=nn)
        retvals[kk] = rstr
        outs.append(rstr)
        printvals.append({"base": base, "conv": conv, "name": nn})

    if print_output:
        print("")

        for val in printvals:
            click.secho('{base:30s}'.format(base=val['base']),
                        bold=True, nl=False, fg=_COLOR_BASE)
            click.secho('{conv:20s}'.format(conv=val['conv']),
                        bold=False, nl=False, fg=_COLOR_CONV)
            click.secho(' : ', bold=True, nl=False)  # , fg="green")
            click.secho('{name}'.format(name=val["name"]),
                        bold=False, nl=True, fg=_COLOR_NAME)

        print("")

    return retvals


def output_api(results):
    """Convert cosmological parameters from numeric values to formatted-strings.

    Arguments
    ---------
    results : `dict`
        Cosmological parameters in numerical form.

    Returns
    -------
    retvals : `dict`
        Cosmological parameters as nicely formatted strings.

    """

    keys = ['z', 'a', 'dc', 'dl',
            'da', 'arc', 'tl',
            'ta', 'dm']
    symbs = ['z', 'a', 'D_c', 'D_L',
             'D_A', 'Arcsec', 'T_lb',
             'T_a', 'DM']
    types = [None, None, 'Mpc', 'Mpc',
             'Mpc', 'pc', 'Gyr',
             'Gyr', None]
    names = ['Redshift', 'Scale-factor', 'Comoving Distance', 'Luminosity Distance',
             'Angular Diameter Distance', 'Arcsecond Scale', 'Lookback Time',
             'Age of the Universe', 'Distance Modulus']

    retvals = {}
    for kk, ss, tt, nn in zip(keys, symbs, types, names):
        vv = results[kk]
        try:
            vv = vv.item()
        except AttributeError:
            pass
        v_std = vv if tt is None else vv.to(tt)
        base = "{:.4f}".format(v_std)
        # try:
        #     base = "{:.4f}".format(v_std)
        # except Exception:
        #     print("Failed constructing `base` string on '{}' ({})".format(nn, ss))
        #     print("  Value = '{}', type = '{}'".format(v_std, type(v_std)))
        #     raise

        retvals[kk] = base

    return retvals


def parse_input(inval, unit=None):
    """Convert an input string value into a number, possibly with units.

    If `unit` is given and the input is a unitless float, then the given unit is added.
    e.g. "2.3e12 cm" will be parsed into an `astropy.units.quantity.Quantity`

    Arguments
    ---------
    inval : str
    unit : `astropy.unit.Unit` or `None`
        Unit to be added to the given input (if it doesn't already specify a unit value).

    Returns
    -------
    val : `astropy.unit.quantity.Quantity`
        Parsed value.  This value will have units attached if either the input string specified
        them, or if a `unit` value was given.

    """
    try:
        val = float(inval)
    except ValueError:
        try:
            val = ap.units.quantity.Quantity(inval)
        except Exception:
            print("__main__.parse_input()")
            print("Failed to convert '{}'".format(inval))
            raise
    else:
        if unit is not None:
            val = val * unit

    return val


def parse_args(args):
    """Initialize desired command line options, retrieve their values.
    """

    parser = argparse.ArgumentParser(description="cosmopy: cosmological calculator.")

    # Target Parameters
    parser.add_argument('-z', type=float, default=None,
                        help='target redshift z')
    parser.add_argument('-a', type=float, default=None,
                        help='target scale factor a')
    parser.add_argument('-dc', '-cd', default=None,
                        help='target coming distance D_C')
    parser.add_argument('-dl', '-ld', default=None,
                        help='target luminosity distance D_L')
    parser.add_argument('-tl', '-lt', default=None,
                        help='target look-back time T_L')
    parser.add_argument('-ta', '-at', default=None,
                        help='target universe age T_A')

    parser.add_argument('-v', '--version', action="store_true", default=False,
                        help='print version information.')

    # print("args = ", args)
    sets = parser.parse_args(args)

    if sets.version:
        from . import __version__
        print("cosmopy :: https://github.com/lzkelley/cosmopy/")
        print(os.path.abspath(__file__))
        print(__version__)
        sys.exit(1)

    arg_given = False
    arg_vars = vars(sets)
    arg_vars.pop('version')
    # print(arg_vars)
    for kk, vv in arg_vars.items():
        if vv is not None:
            arg_given = True

    if not arg_given:
        print("No argument given.  Must provide an input argument.\n")
        parser.print_help()
        print("")
        sys.exit(1)

    return sets


def api(key, val, cosmo=None):
    """Primary method for external API access to this package.

    Arguments
    ---------
    key : str
        Specification of target cosmological parameter.  Must be one of `_RESULTS_PARS`.
    val : str or scalar
        The value of the target cosmological parameter.
        Can be a scalar value in CGS units, or a string including unit specification.

    Returns
    -------
    retvals : dict
        Dictionary of cosmological parameters as formatted strings.

    """
    sets = {kk: None for kk in _RESULTS_PARS}
    if key not in sets:
        print("Valid keys: '{}'".format(_RESULTS_PARS))
        raise KeyError("Unrecognized key '{}'".format(key))

    # Set the given value
    sets[key] = val
    # Convert to `Namespace` object (expected by the `calc_basic` method)
    sets = argparse.Namespace(**sets)

    # Load cosmology object if needed
    if cosmo is None:
        cosmo = get_cosmology()

    # Calculate
    # ----------------
    # Calculate basic cosmological values
    results = calc_basic(cosmo, sets)
    # Calculate additional derived values
    results = calc_derived(results)
    # Format output string and return as dictionary
    retvals = output_api(results)

    return retvals


def get_cosmology(*args, **kwargs):
    """Construct an instance of the `Cosmology` class and return.
    """
    cosmo = Cosmology(*args, **kwargs)
    return cosmo


def main(args=None):
    """Primary computational method for command-line access to calculations.

    Loads command line arguments and a `Cosmology` instance.
    Calculates cosmological distance measures.
    Formats and prints the output to stdout.

    """
    # Initialize
    # ----------------
    # Load arguments
    if args is None:
        args = sys.argv[1:]
    sets = parse_args(args)
    # Load cosmology object
    cosmo = get_cosmology()

    # Calculate
    # ----------------
    # Calculate basic cosmological values
    results = calc_basic(cosmo, sets)
    # Calculate additional derived values
    results = calc_derived(results)
    # Format and print outputs
    output_print(results)

    return


if __name__ == "__main__":
    main()



"""Test methods for `cosmopy.cosmology.py`.
"""

import numpy as np
import astropy as ap
import astropy.units

import pytest


def _stats(vv, percs=[0, 25, 50, 75, 100]):
    ss = np.percentile(vv, percs)
    msg = ", ".join(['{:.4e}'.format(_s) for _s in ss])
    return msg


class TestCosmology:

    funcs_forw = ["luminosity_distance", "comoving_distance", "age", "lookback_time"]
    funcs_back = ["dlum_to_z", "dcom_to_z", "tage_to_z", "tlbk_to_z"]
    z_low = 0.001
    vals_low = [1.3353e+25, 1.3340e+25, 4.3356e+17, 4.4475e+14]

    def test_init(self):
        from cosmopy.cosmology import Cosmology
        cosmo = Cosmology()
        print(cosmo)

        # Cannot provide both `h` and `H0`
        with pytest.raises(ValueError):
            Cosmology(h=0.697, H0=2.24e-18)

        return

    def test__init_interp_grid(self):
        from cosmopy.cosmology import Cosmology

        zpnts = [10.0, 1.0]
        numz = 10

        grid = Cosmology._init_interp_grid(zpnts, numz)
        assert grid.size == numz * len(zpnts)

        with pytest.raises(ValueError):
            Cosmology._init_interp_grid(zpnts[::-1], numz)

        return

    def test_interp_grid(self):
        from cosmopy.cosmology import Cosmology
        cosmo = Cosmology()
        grid_z = cosmo._grid_z
        num_z = grid_z.size
        print("z-grid has size: {}".format(num_z))
        assert np.isclose(grid_z[-1], 0.0)

        print("cosmo._grid_a.size, num_z = {}, {}".format(cosmo._grid_a.size, num_z))
        assert (cosmo._grid_a.size == num_z)
        print("cosmo._sort_z.size, num_z = {}, {}".format(cosmo._sort_z.size, num_z))
        assert (cosmo._sort_z.size == num_z)

        dz = np.diff(grid_z[cosmo._sort_z])
        assert (np.all(dz > 0.0))

        return

    def test_interp(self):
        from cosmopy.cosmology import Cosmology

        xx = np.array([1.0, 2.0])
        yy = np.array([10.0, 20.0])
        val = 15.0

        zz = Cosmology._interp(1.5, xx, yy)
        print("Interp = {}, should be = {}".format(zz, val))
        assert np.isclose(zz, val)

        zz = Cosmology._interp(1.5, xx[::-1], yy[::-1])
        assert np.isclose(zz, val)

        # Out of domain should be NaN
        bad = Cosmology._interp(0.5, xx, yy)
        assert np.isnan(bad)

        return

    def test_a_z_conv(self):
        from cosmopy.cosmology import Cosmology
        # These are static-methods so we dont need an instance

        zz = [3.0, 1.0, 0.5, 0.0]
        aa = [0.25, 0.5, 0.666666666, 1.0]

        for z, a in zip(zz, aa):
            _a = Cosmology.z_to_a(z)
            _z = Cosmology.a_to_z(a)

            print("z = {} ===> a = {} ({})".format(z, _a, a))
            assert np.isclose(z, _z)

            print("a = {} ===> z = {} ({})".format(a, _z, z))
            assert np.isclose(a, _a)

        # Test array input
        assert (np.allclose(Cosmology.z_to_a(zz), aa))
        assert (np.allclose(Cosmology.a_to_z(aa), zz))

        # Test out-of-domain errors
        with pytest.raises(ValueError):
            Cosmology.a_to_z(-0.1)
        with pytest.raises(ValueError):
            Cosmology.a_to_z(1.1)
        with pytest.raises(ValueError):
            Cosmology.z_to_a(-0.2)

        return

    def test_forward(self):
        """Test "forward" conversion from redshift to other parameters
        """
        from cosmopy.cosmology import Cosmology
        cosmo = Cosmology()

        z = self.z_low
        for vv, ff in zip(self.vals_low, self.funcs_forw):
            _v = getattr(cosmo, ff)(z).cgs.value
            exp = np.power(10.0, np.floor(np.log10(vv)))
            print("{}({}) = {} (should be: {})".format(ff, z, _v, vv))
            assert np.isclose(_v/exp, vv/exp, 1e-3)

        return

    def test_backward(self):
        """Test "backward" conversion from parameters to redshift
        """
        from cosmopy.cosmology import Cosmology
        cosmo = Cosmology()

        z = self.z_low
        for vv, ff in zip(self.vals_low, self.funcs_back):
            _z = getattr(cosmo, ff)(vv)
            print("{}({}) = {} (should be: {})".format(ff, vv, _z, z))
            assert np.isclose(z, _z, atol=1e-3)

        return

    def test__get_grid(self):
        from cosmopy.cosmology import Cosmology
        cosmo = Cosmology()

        grid, names, units = cosmo._get_grid()
        z_grid = cosmo._grid_z
        num = z_grid.size
        print(names)
        print(units)
        assert grid.shape == (num, 6,)
        assert np.allclose(grid[:, 0], cosmo._grid_z)

        ind = names.index('dl')
        dl = cosmo._grid_dlum
        un = ap.units.Unit(units[ind])
        print("'dl' ind = {}, unit = {}".format(ind, un))
        grid_dl = grid[:, ind] * un
        grid_dl = grid_dl.cgs.value
        print("dl ind = {}, {}, {}".format(ind, dl[:4], grid_dl[:4]))
        assert np.allclose(grid_dl, dl)

    def test_dvc_dz(self):
        from cosmopy.cosmology import Cosmology
        cosmo = Cosmology()

        NUM = 1000
        RTOL = 1e-6

        redz = 10**np.random.uniform(-2, 1, int(NUM))
        redz = np.sort(redz)

        vals = cosmo.dVcdz(redz, cgs=True)
        check = cosmo.differential_comoving_volume(redz).cgs.value * 4*np.pi
        print("redz  : ", _stats(redz))
        print("dVcdz : ", _stats(vals))
        print("check : ", _stats(check))
        assert np.allclose(vals, check, rtol=RTOL), "Error in dvcdz exceeds rtol = {}".format(RTOL)

        # use CGS units
        print("cgs = True")
        vals = cosmo.dVcdz(redz, cgs=True)
        check = cosmo.differential_comoving_volume(redz).cgs.value * 4*np.pi
        print("\ncheck = ", check[:4], check[-4:])

        print("redz  : ", _stats(redz))
        print("dVcdz : ", _stats(vals))
        print("check : ", _stats(check))
        assert np.allclose(vals, check, rtol=RTOL), "Error in dvcdz(cgs=True) exceeds rtol = {}".format(RTOL)

        # use standard units
        print("cgs = False")
        vals = cosmo.dVcdz(redz, cgs=False).value
        check = cosmo.differential_comoving_volume(redz).value * 4*np.pi
        print("\ncheck = ", check[:4], check[-4:])

        print("redz  : ", _stats(redz))
        print("dVcdz : ", _stats(vals))
        print("check : ", _stats(check))
        assert np.allclose(vals, check, rtol=RTOL), "Error in dvcdz(cgs=False) exceeds rtol = {}".format(RTOL)

        return

    def test_dtdz(self):
        from cosmopy.cosmology import Cosmology
        cosmo = Cosmology()

        NUM = 1000
        redz = 10 ** np.random.uniform(-2, 1, int(NUM))
        redz = np.sort(redz)

        def check(zz):
            zz = np.asarray(zz)
            efac = cosmo.efunc(zz)
            th = cosmo.hubble_time.to('s').value
            rv = th / (1.0 + zz) / efac
            return rv

        vals = cosmo.dtdz(redz)
        check = check(redz)
        assert np.allclose(vals, check)

        # precomputed values for [Omega0 = 0.2880, OmegaBaryon = 0.0472, HubbleParam = 0.6933]
        test = [0.001, 0.01, 0.1, 0.5, 1.0, 5.0, 10.0]
        check = [4.4443e+17, 4.3875e+17, 3.8660e+17, 2.2865e+17, 1.2814e+17, 9.3516e+15, 2.0647e+15]
        vv = cosmo.dtdz(test)
        assert np.allclose(vv, check, rtol=1e-4)

        return


def _get_err(aa, bb):
    err = (aa - bb) / np.minimum(aa, bb)
    err = np.fabs(err)
    return err.mean(), err.max()


def test_z_to__methods():
    from cosmopy.cosmology import Cosmology
    cosmo = Cosmology()

    NUM = 1000
    AVE_THRESH = 1e-6
    MAX_THRESH = 1e-4

    redz = 10**np.random.uniform(-2, 1, int(NUM))
    redz = np.sort(redz)

    funcs = ['tage', 'tlbk', 'dlum', 'dcom']
    funcs_check = ['age', 'lookback_time', 'luminosity_distance', 'comoving_distance']
    funcs_for = ["z_to_" + ff for ff in funcs]
    funcs_back = [ff + "_to_z" for ff in funcs]

    for ii, (func_for, check, func_back) in enumerate(zip(funcs_for, funcs_check, funcs_back)):
        _err = "{} :: Failed - ".format(funcs[ii])

        val_for = getattr(cosmo, func_for)(redz)
        chk = getattr(cosmo, check)(redz).cgs.value
        val_back = getattr(cosmo, func_back)(val_for)

        chk_ave, chk_max = _get_err(val_for, chk)
        print("forward errors: ave={:.4e}, max={:.4e}".format(chk_ave, chk_max))

        # --- Convert forwards, and check against astropy default calculation
        err = _err + "forward error exceeded average threshold {} > {}!".format(chk_ave, AVE_THRESH)
        assert chk_ave < AVE_THRESH, err

        err = _err + "forward error exceeded maximum threshold {} > {}!".format(chk_max, MAX_THRESH)
        assert chk_max < MAX_THRESH, err

        # --- Convert backwards, and check against input redshifts
        bck_ave, bck_max = _get_err(val_back, redz)
        print("backward errors: ave={:.4e}, max={:.4e}".format(bck_ave, bck_max))

        err = _err + "backward error exceeded average threshold {} > {}!".format(bck_ave, AVE_THRESH)
        assert bck_ave < AVE_THRESH, err

        err = _err + "backward error exceeded maximum threshold {} > {}!".format(bck_max, MAX_THRESH)
        assert bck_max < MAX_THRESH, err

    return


def test_a_to__methods():
    from cosmopy.cosmology import Cosmology
    cosmo = Cosmology()

    NUM = 1000
    AVE_THRESH = 1e-6
    MAX_THRESH = 1e-4

    _redz = 10**np.random.uniform(-2, 1, int(NUM))
    _redz = np.sort(_redz)
    scafa = cosmo.z_to_a(_redz)

    funcs = ['tage', 'tlbk', 'dlum', 'dcom']
    funcs_check = ['age', 'lookback_time', 'luminosity_distance', 'comoving_distance']
    funcs_for = ["a_to_" + ff for ff in funcs]
    funcs_back = [ff + "_to_a" for ff in funcs]

    for ii, (func_for, check, func_back) in enumerate(zip(funcs_for, funcs_check, funcs_back)):
        _err = "{} :: Failed - ".format(funcs[ii])

        val_for = getattr(cosmo, func_for)(scafa)
        chk = getattr(cosmo, check)(_redz).cgs.value
        val_back = getattr(cosmo, func_back)(val_for)
        print(funcs[ii], "scafa    : ", _stats(scafa))
        print(funcs[ii], "redz     : ", _stats(_redz))
        print(funcs[ii], "forward  : ", _stats(val_for))
        print(funcs[ii], "check    : ", _stats(chk))
        print(funcs[ii], "backward : ", _stats(val_back))

        chk_ave, chk_max = _get_err(val_for, chk)
        print("{} forward errors: ave={:.4e}, max={:.4e}".format(funcs[ii], chk_ave, chk_max))

        # --- Convert forwards, and check against astropy default calculation
        err = _err + "forward error exceeded average threshold {} > {}!".format(chk_ave, AVE_THRESH)
        assert chk_ave < AVE_THRESH, err

        err = _err + "forward error exceeded maximum threshold {} > {}!".format(chk_max, MAX_THRESH)
        assert chk_max < MAX_THRESH, err

        # --- Convert backwards, and check against input redshifts
        bck_ave, bck_max = _get_err(val_back, scafa)
        print("{} backward errors: ave={:.4e}, max={:.4e}".format(funcs[ii], bck_ave, bck_max))

        err = _err + "backward error exceeded average threshold {} > {}!".format(bck_ave, AVE_THRESH)
        assert bck_ave < AVE_THRESH, err

        err = _err + "backward error exceeded maximum threshold {} > {}!".format(bck_max, MAX_THRESH)
        assert bck_max < MAX_THRESH, err

    return



"""
"""

__version__ = "3.2.1"
__author__ = "Luke Zoltan Kelley <lzkelley@northwestern.edu>"
__license__ = "MIT"

import astropy as ap
import astropy.constants  # noqa
import astropy.units  # noqa


NWTG = ap.constants.G.cgs.value
SPLC = ap.constants.c.cgs.value
MSOL = ap.constants.M_sun.cgs.value
LSOL = ap.constants.L_sun.cgs.value
RSOL = ap.constants.R_sun.cgs.value
PC = ap.constants.pc.cgs.value
AU = ap.constants.au.cgs.value

ARCSEC = ap.units.arcsec.cgs.scale       # arcsecond in radians
YR = ap.units.year.to(ap.units.s)       # year in seconds

U_SEC = ap.units.Unit('s')       # second unit
U_CM = ap.units.Unit('cm')       # centimeter unit

from . cosmology import Cosmology          # noqa
from . __main__ import get_cosmology, api  # noqa



"""Test methods for `cosmopy.__main__.py`.
"""

import runpy

import numpy as np
import astropy as ap
import astropy.units

import pytest


class TestMain:

    np.random.seed(986523)
    z_low = 0.001
    vals_low = [1.3353e+25, 1.3340e+25, 4.3356e+17, 4.4475e+14]
    arg_keys = ['dl', 'dc', 'ta', 'tl']

    def test__module(self):
        with pytest.raises(SystemExit):
            runpy.run_module('cosmopy', run_name='__main__')

        # globs = {'sys.argv': ['cosmopy', '-z', '0.1']}
        # print("\n\n\n=================================================\n\n\n")
        # runpy.run_module('cosmopy', run_name='__main__', init_globals=globs)
        return

    def test__parse_args(self):
        from cosmopy.__main__ import parse_args

        with pytest.raises(SystemExit):
            parse_args([])

        with pytest.raises(SystemExit):
            parse_args(['-v'])

        vals = self.vals_low + [0.1, 0.5]
        keys = self.arg_keys + ['z', 'a']

        for vv, kk in zip(vals, keys):
            args = ["-" + kk, "{:.4e}".format(vv)]
            print("args = ", args)
            parse_args(args)

        return

    def test__main(self):
        from cosmopy.__main__ import main

        with pytest.raises(SystemExit):
            main([])

        with pytest.raises(SystemExit):
            main(['-v'])

        keys = self.arg_keys + ['z', 'a']
        vals = self.vals_low + [0.1, 0.5]

        for kk, vv in zip(keys, vals):
            args = ["-" + kk, "{:.4e}".format(vv)]
            print("args = ", args)
            main(args)

        return

    def test__parse_input(self):
        from cosmopy.__main__ import parse_input
        from astropy.units import Unit

        ins = ['-12', '+36', '2.2', '2.3e23', '-1.98e-1', 2.3]
        outs = [-12, 36, 2.2, 2.3e23, -1.98e-1, 2.3]

        for ii, oo in zip(ins, outs):
            print("In: '{}'".format(ii))
            rv = parse_input(ii)
            print("\tOut: '{}' (vs. '{}')".format(rv, oo))
            assert np.isclose(rv, oo)

        ins = ['2.3e12', '7.4e2']
        dunits = [Unit('s'), Unit('yr')]
        outs = [2.3e12, 23352624000.0]

        for ii, dd, oo in zip(ins, dunits, outs):
            print("In: '{}'".format(ii))
            rv = parse_input(ii, dd).cgs.value
            print("\tOut: '{}' (vs. '{}')".format(rv, oo))
            assert np.isclose(rv, oo)

        ins = ['-35 kg', '7.4e2 s', '2.2 cm', '2.3e23 yr', '-1.98e-1 pc']
        outs = [-35e3, 7.4e2, 2.2, 7.258248e+30, -6.10964161e+17]

        for ii, oo in zip(ins, outs):
            print("In: '{}'".format(ii))
            rv = parse_input(ii).cgs.value
            print("\tOut: '{}' (vs. '{}')".format(rv, oo))
            assert np.isclose(rv, oo, atol=np.fabs(1e-6*oo))

        # Should fail with unrecognized units
        with pytest.raises(ValueError):
            parse_input('12 Monkeys')

        return

    def test__get_cosmology(self):
        from cosmopy.__main__ import get_cosmology

        cosmo = get_cosmology()
        print(cosmo)

        dl = cosmo.luminosity_distance(self.z_low).cgs.value
        print(dl, self.vals_low[0])
        assert np.isclose(dl, self.vals_low[0], rtol=1e-3)

        z = cosmo.dlum_to_z(dl)
        print(z, self.z_low)
        assert np.isclose(z, self.z_low)
        return

    def test__api(self):
        from cosmopy.__main__ import api

        # Use a key that shouldn't be recognized
        with pytest.raises(KeyError):
            api('m', 0.2)

        # These values were calculated using built-in astropy cosmology instance, for comparison
        outs = {
            'z': '0.1000', 'a': '0.9091', 'ta': '12.4377 Gyr', 'dl': '465.1654 Mpc',
            'da': '384.4342 Mpc', 'tl': '1.3150 Gyr', 'dc': '422.8776 Mpc',
            'arc': '1863.7897 pc', 'dm': '38.3380'
        }

        def compare(aa):
            rv = api(*aa)
            tests = []
            print("input args = ", args)
            for kk, v1 in outs.items():
                v2 = rv[kk]
                q1 = ap.units.Quantity(v1).cgs.value
                q2 = ap.units.Quantity(v2).cgs.value
                tt = np.isclose(q1, q2, rtol=1e-3)
                print("{} : {} vs {} :: {}".format(kk, v1, v2, tt))
                tests.append(tt)

            assert np.all(tests)

        args = ['z', '0.1']
        compare(args)

        args = ['dc', '423.1241Mpc']
        compare(args)

        args = ['dc', '0.4231241Gpc']
        compare(args)

        return

    def test__calc_basic(self):
        from cosmopy.__main__ import calc_basic, _RESULTS_PARS, get_cosmology
        import argparse

        cosmo = get_cosmology()

        # Should raise error with no settings set
        sets = {kk: None for kk in _RESULTS_PARS}
        args = argparse.Namespace(**sets)
        with pytest.raises(ValueError):
            calc_basic(cosmo, args)

        return



from setuptools import setup

FNAME_README = "README.rst"
FNAME_REQUIREMENTS = "requirements.txt"
FNAME_VERSION = "cosmopy/VERSION.txt"

with open(FNAME_REQUIREMENTS, "r") as inn:
    requirements = inn.read().splitlines()

with open(FNAME_README, "r") as inn:
    long_description = inn.read().strip()

with open(FNAME_VERSION, "r") as inn:
    version = inn.read().strip()

setup(
    name="cosmopy",
    author="Luke Zoltan Kelley",
    author_email="lzkelley@northwestern.edu",
    version=version,
    description="",
    license="MIT",
    url="https://github.com/lzkelley/cosmopy",
    download_url="https://github.com/lzkelley/cosmopy/archive/v{}.tar.gz".format(version),
    packages=['cosmopy'],
    entry_points={
        'console_scripts': [
            'cosmo = cosmopy.__main__:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['utilities', 'physics', 'astronomy', 'cosmology', 'astrophysics', 'calculator'],
    python_requires=">=3.5",
)
